#include<stdlib.h>
#include<stdio.h>

#define INT int
int FILE_err=0;
INT *item_order=NULL, *item_perm=NULL, *org_frq=NULL;

typedef struct {
  INT **trsact, *buf, *w;   // transactions, its buffer, and their weights
  INT trsact_num, eles;  // #transactions, total #elements
} TRSACT;

typedef struct {
  INT trsact_num, item_max;  // #transactions in original database, #items
  INT **occ, *occ_buf, *occt;  // occurrences, its buffer, and their sizes
  INT *frq;    // array for frequency of each item, and those in the original database
  INT *itemset, itemset_siz;  // itemset array and its size
  INT *add, add_siz;
  INT *jump, jump_siz; // candidate queue and its size
  INT sigma; // minimum support
  
  INT *head, *nxt, *mark;  // variables for shrinking the database
} LCM;

/* allocate memory with error routine */
char *alloc_memory (size_t siz){
  char *p = (char *)calloc (siz, 1);
  if ( p == NULL ){
    printf ("out of memory\n");
    exit (1);
  }
  return (p);
}

/* re-allocate memory and fill the newly allocated part by 0, with error routine */
char *realloc_memory (char *p, int sizof, size_t siz, size_t *org_siz){
  size_t i;
  p = (char *)realloc(p, siz*sizof);
  if ( p == NULL ){
    printf ("out of memory\n");
    exit (1);
  }
  for ( i=(*org_siz)*sizof ; i<siz*sizof ; i++ ) p[i] = 0;
  *org_siz = siz;
  return (p);
}

/* comparison subroutines for quick-sort */
int qsort_cmp_frq (const void *x, const void *y){
  if ( org_frq[*((INT *)x)] > org_frq[*((INT *)y)] ) return (-1);
  else return ( org_frq[*((INT *)x)] < org_frq[*((INT *)y)] );
}
int qsort_cmp_idx (const void *x, const void *y){
  if ( item_order[*((INT *)x)] < item_order[*((INT *)y)] ) return (-1);
  else return ( item_order[*((INT *)x)] > item_order[*((INT *)y)] );
}
int qsort_cmp__idx (const void *x, const void *y){
  if ( item_order[*((INT *)x)] > item_order[*((INT *)y)] ) return (-1);
  else return ( item_order[*((INT *)x)] < item_order[*((INT *)y)] );
}


/* read an integer from the file */
INT FILE_read_int (FILE *fp){
  INT item;
  int flag =1;
  int ch;
  FILE_err = 0;
  do {
    ch = fgetc (fp);
    if ( ch == '\n' ){ FILE_err = 5; return (0); }
    if ( ch < 0 ){ FILE_err = 6; return (0); }
    if ( ch=='-' ) flag = -1;
  } while ( ch<'0' || ch>'9' );
  for ( item=(int)(ch-'0') ; 1 ; item=item*10 +(int)(ch-'0') ){
    ch = fgetc (fp);
    if ( ch == '\n' ){ FILE_err = 1; return (flag*item); }
    if ( ch < 0 ){ FILE_err = 2; return (flag*item); }
    if ( (ch < '0') || (ch > '9')) return (flag*item);
  }
}

void TRSACT_print (LCM *L, TRSACT *T, INT item){
  INT i, j, tID;
  printf ("------------ end_mark = %d\n", item);
  for ( tID=0 ; tID<T->trsact_num ; tID++ ){
    printf ("%d::  ", tID);
    for ( j=0 ; (i=T->trsact[tID][j]) != item ; j++ ) printf ("%d ", i);
    printf (" (%d)\n", T->w[tID]);
  }
  printf ("============\n");
}

/* load a transaction from the input file to memory */
INT TRSACT_file_load (TRSACT *T, char *fname){
  INT item, t, item_max=0;
  size_t cnt=0, new_cnt;
  FILE *fp = fopen (fname,"r");

  if ( !fp ){ printf ("file open error\n"); exit (1); }
  T->trsact_num = 0; // set #transactions to 0
  T->eles = 0;

   // scan file to count the 
  do {
    do {
      item = (INT)FILE_read_int (fp);
      if ( (FILE_err&4) == 0){  // got an item from the file before reaching to line-end
        if ( item >= (INT)cnt ){
          new_cnt = cnt*2; if ( item >= (INT)new_cnt ) new_cnt = item +1;
          org_frq = (INT*)realloc_memory((char*)org_frq, sizeof(INT), new_cnt, &cnt);
        }
        if ( item >= item_max ) item_max = item+1;  // uppdate the maximum item
        org_frq[item]++;
        T->eles ++;
      }
    } while ((FILE_err&3)==0);
    T->trsact_num++;  // increase #transaction
  } while ( (FILE_err&2)==0);
  
    // compute the item ordering
  item_order = (INT *)alloc_memory ( sizeof(INT) * item_max );
  item_perm = (INT *)alloc_memory ( sizeof(INT) * item_max );
  for ( t=0 ; t<item_max ; t++ ) item_perm[t] = t;
  qsort (item_perm, item_max, sizeof(INT), qsort_cmp_frq);
  for ( t=0 ; t<item_max ; t++ ) item_order[item_perm[t]] = t;

    // allocate memory for transactions
  T->buf = (INT *)alloc_memory ( sizeof(INT) * (T->eles+T->trsact_num) );
  T->trsact = (INT **)alloc_memory ( sizeof(INT*) * T->trsact_num );
  T->w = (INT *)alloc_memory ( sizeof(INT) * T->trsact_num );

    // load the file
  cnt = 0;
  fseek (fp, 0, SEEK_SET);
  for ( t=0 ; t<T->trsact_num ; t++){
    T->trsact[t] = &T->buf[cnt];
    do {
      item = FILE_read_int (fp);
      if ( (FILE_err&4) == 0){ T->buf[cnt] = item; cnt++;}
    } while ((FILE_err&3)==0);
    qsort (T->trsact[t], &T->buf[cnt] - T->trsact[t], sizeof(INT), qsort_cmp_idx); // sort each transaction
    T->buf[cnt] = item_max;  // attach an end mark
    T->w[t] = 1;
    cnt++;
  }
  return (item_max);
}

/* count the frequency of each item */
void frq_count (LCM *L, TRSACT *T, int item){
  INT t, i, j, tID;
  for ( t=0 ; t<L->occt[item] ; t++ ){
    tID = L->occ[item][t];
    for ( j=0 ; (i=T->trsact[tID][j]) != item ; j++ ){
      if ( L->frq[i] == 0 ) L->jump[L->jump_siz++] = i;
      L->frq[i] += T->w[tID];
      L->occt[i]++;
    }
  }
}
/*
void frq_count (LCM *L, TRSACT *T, int item){
  INT *t, *i;
  for ( t=L->occ[item] ; t<L->occ[item] + L->occt[item] ; t++ ){
    for ( i = T->trsact[*t] ; *i != item ; i++ ){
      if ( L->frq[*i] == 0 ) L->jump[L->jump_siz++] = *i;
      L->frq[*i] += T->w[*t];
      L->occt[*i]++;
    }
  }
}
*/

/* remove infrequent items from the candidate list, and sort it */
size_t rm_infrq (LCM *L, INT jt, INT frq){
  INT i, ii, item;
  size_t cnt=0;
  for ( i=ii=jt ; i<L->jump_siz ; i++ ){
    item = L->jump[i];
    if ( L->frq[item] == frq ){
      L->add[L->add_siz++] = item;
      L->frq[item] = 0;
    } else if ( L->frq[item] >= L->sigma ){
      L->jump[ii++] = item;
      cnt += L->occt[item];
    } else {
      L->frq[item] = 0;
    }
    L->occt[item] = 0;
  }
  L->jump_siz = ii;
  qsort (&L->jump[jt], L->jump_siz-jt, sizeof(INT), qsort_cmp__idx);  // sort the items 
  return (cnt);
}


/* compute the occurrences of all items by occurrence deliver */
void occ_deliv (LCM *L, TRSACT *T, int item){
  INT i, j, tID;
  for ( tID=0 ; tID<T->trsact_num ; tID++ ){
    for ( j=0 ; (i=T->trsact[tID][j]) != item ; j++ ){
      L->occ[i][L->occt[i]++] = tID;
    }
  }
}
/*
void occ_deliv (LCM *L, TRSACT *T, int item){
  INT tID, *i;
  for ( tID=0 ; tID<T->trsact_num ; tID++ ){
    for ( i = T->trsact[tID] ; *i != item ; i++ ){
      L->occ[*i][L->occt[*i]++] = tID;
    }
  }
}

/* construct a conditional database from occ */
TRSACT condi_dat (LCM *L, TRSACT *T, int item, size_t cnt){
  INT t, i, j, tID;
  TRSACT TT;
  
  TT.buf = (INT *)alloc_memory ( sizeof(INT) * (cnt +L->occt[item]) );
  TT.trsact = (INT **)alloc_memory ( sizeof(INT*) * L->occt[item] );
  TT.w = (INT *)alloc_memory ( sizeof(INT) * L->occt[item] );
  TT.trsact_num = L->occt[item];
  cnt = 0;
  for ( t=0 ; t<L->occt[item] ; t++ ){
    TT.trsact[t] = &TT.buf[cnt];
    tID = L->occ[item][t];
    TT.w[t] = T->w[tID];
    for ( j=0 ; (i=T->trsact[tID][j]) != item ; j++ ){
      if ( L->frq[i] >= L->sigma ) TT.buf[cnt++] = i;
    }
    TT.buf[cnt++] = item;
  }
  return (TT);
}

/* shrink the database by unifying the identical transactions */
void shrink_dat (LCM *L, TRSACT *T, int item){
  INT t, i, j=0, h, tID, thead=0, thead2, thead_, it=L->itemset_siz;
    // left space of L->itemset will be used to memory accessed items.
    // itemset never overflow, since itemset is composed of items smaller than
    // "item" (in the term of item_ordering),
    // and in this routine we access items larger than "item"

    // make initial list containing all transactions
  for ( tID=0 ; tID<T->trsact_num ; tID++ ){
    L->nxt[tID] = tID + 1;
    L->mark[tID] = 0;
  }
  L->nxt[T->trsact_num-1] = L->trsact_num;
  
  while (1){
    thead2 = L->trsact_num;   // make an empty list
    if ( thead == L->trsact_num ) break;
    
    while (thead < L->trsact_num){

      i = T->trsact[thead][j];
      h = L->head[i];
      thead_ = L->nxt[thead];
      if ( i == item ){  // end of transaction
        if ( h != L->trsact_num && L->mark[h] == L->mark[thead] ){
          T->w[h] += T->w[thead];
          L->mark[thead] = L->trsact_num;  // put deleted mark
         } else L->head[i] = thead;
        // not the end of transaction
      } else if ( h == L->trsact_num ){ // new group will be made
        L->nxt[thead] = thead2;
        thead2 = thead;
        L->head[i] = thead;
        L->itemset[L->itemset_siz++] = i;
      } else if ( L->mark[h] != L->mark[thead] ){  // replace the group
        L->nxt[thead] = thead2;
        thead2 = thead;
        L->head[i] = thead;
        L->mark[h] = h;     // put mark to previous head
      } else {    // member of the group
        L->nxt[thead] = L->nxt[h];
        L->nxt[h] = thead;
        L->mark[thead] = h;  // put group mark
      }
      thead = thead_;
    }

    j++;
    thead = thead2;

      // clear the written heads
    while (L->itemset_siz > it ){
      h = L->itemset[--L->itemset_siz];
      L->mark[L->head[h]] = L->head[h];
      L->head[h] = L->trsact_num;  // put mark to previous head
    }
    L->head[item] = L->trsact_num;
  }

    // shift the remaining transactions
  for ( t=tID=0 ; tID<T->trsact_num ; tID++ ){
    if ( L->mark[tID] != L->trsact_num ){
      T->trsact[t] = T->trsact[tID];
      T->w[t++] = T->w[tID];
    }
  }
  T->trsact_num = t;

}

/* output an itemset */
void output_itemset (LCM *L, INT frq, INT addt){
  INT i;
  if ( addt == 0 ){
    for ( i=0 ; i<L->itemset_siz ; i++) printf ("%d ", L->itemset[i]);
    printf ("(%d)\n", frq);
  } else {
    addt--;
    output_itemset (L, frq, addt);
    L->itemset[L->itemset_siz++] = L->add[addt];
    output_itemset (L, frq, addt);
    L->itemset_siz--;
    addt++;
  }
}

/* main recursive call */
void LCMmain (LCM *L, TRSACT *T, INT item){
  INT i, jt = L->jump_siz, at = L->add_siz, flag=0;
  TRSACT TT;
  size_t cnt;

  frq_count (L, T, item);
  cnt = rm_infrq (L, jt, L->frq[item]);
  output_itemset (L, L->frq[item], L->add_siz);  // output an itemset
  TT = condi_dat (L, T, item, cnt); // make conditional database
  
  if ( L->jump_siz - jt > 2 && TT.trsact_num > 10 ) shrink_dat (L, &TT, item);

  occ_deliv (L, &TT, item);  // compute each item frequency

  while ( L->jump_siz > jt ){
    i = L->jump[--L->jump_siz];
    if ( L->frq[i] >= L->sigma ){
      L->itemset[L->itemset_siz++] = i;
//      if ( flag == 0 ){ flag = 1; output_itemset (L, L->frq[i], L->add_siz); }
//      else {
        LCMmain (L, &TT, i);
//      }
      L->itemset_siz--;
    }
    L->occt[i] = 0; L->frq[i] = 0;
  }
  free (TT.trsact);
  free (TT.w);
  free (TT.buf);
  L->add_siz = at;
}



/* main function, for initialization */
main (int argc, char *argv[]){
  TRSACT T;
  LCM L;
  INT i;
  size_t cnt;
  
  if ( argc < 3 ){
    printf ("LCM: enumerate all frequent itemsets\n lcm filename support outputfile\n");
    exit(1);
  }
  L.sigma = atoi(argv[2]);
  L.item_max = TRSACT_file_load (&T, argv[1]);
  L.trsact_num = T.trsact_num;

    // initialize itemset
  L.itemset = (INT *)alloc_memory (sizeof(INT)*L.item_max);
  L.itemset_siz = 0;
  L.add = (INT *)alloc_memory (sizeof(INT)*L.item_max);
  L.add_siz = 0;

    // initialize occ's
  L.occ_buf = (INT *)alloc_memory ( sizeof(INT) * (T.eles+T.trsact_num) );
  L.occ = (INT **)alloc_memory ( sizeof(INT*) * (L.item_max+1) );
  L.occt = (INT *)alloc_memory ( sizeof(INT) * (L.item_max+1) );
  L.frq = (INT *)alloc_memory ( sizeof(INT) * (L.item_max+1) );
  for ( i=0,cnt=0 ; i<L.item_max ; i++ ){
    L.occ[i] = &L.occ_buf[cnt];
    cnt += org_frq[i];
  }
  L.occ[L.item_max] = &L.occ_buf[cnt];
  for ( i=0 ; i<T.trsact_num ; i++ ) L.occ[L.item_max][i] = i;
  L.occt[L.item_max] = L.frq[L.item_max] = T.trsact_num;
  free (org_frq);

    // initialize candidate list
  L.jump = (INT *)alloc_memory ( sizeof(INT) * L.item_max );
  L.jump_siz = 0;

    // initialize variables for database reduction
  L.head = (INT *)alloc_memory ( sizeof(INT) * (L.item_max+1) );
  L.nxt = (INT *)alloc_memory ( sizeof(INT) * T.trsact_num );
  L.mark = (INT *)alloc_memory ( sizeof(INT) * T.trsact_num );
  for ( i=0 ; i<=L.item_max ; i++ ) L.head[i] = L.trsact_num;
  
    // main recursive call
  LCMmain (&L, &T, L.item_max);

    // free the memory
  free (T.trsact);
  free (T.buf);
  free (T.w);

  free (L.jump);

  free (L.occ);
  free (L.occt);
  free (L.occ_buf);
  free (L.frq);

  free (L.itemset);
  free (L.add);

  free (L.mark);
  free (L.nxt);
  free (L.head);
  
  free (item_order);
  free (item_perm);
}



